import os
import math
from itertools import combinations
import numpy as np
from scipy.spatial import KDTree
import logging


class GPSFilter:
    """过滤密集点及孤立点"""

    def __init__(self, output_dir):
        self.log_file = os.path.join(output_dir, 'del_imgs.txt')
        self.logger = logging.getLogger('UAV_Preprocess.GPSFilter')

    @staticmethod
    def _haversine(lat1, lon1, lat2, lon2):
        """计算两点之间的地理距离（单位：米）"""
        R = 6371000  # 地球平均半径，单位：米
        phi1, phi2 = math.radians(lat1), math.radians(lat2)
        delta_phi = math.radians(lat2 - lat1)
        delta_lambda = math.radians(lon2 - lon1)

        a = math.sin(delta_phi / 2) ** 2 + math.cos(phi1) * \
            math.cos(phi2) * math.sin(delta_lambda / 2) ** 2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        return R * c

    @staticmethod
    def _assign_to_grid(lat, lon, grid_size, min_lat, min_lon):
        """根据经纬度和网格大小，将点分配到网格"""
        grid_x = int((lat - min_lat) // grid_size)
        grid_y = int((lon - min_lon) // grid_size)
        return grid_x, grid_y

    def _get_distances(self, points, grid_size):
        """读取图片 GPS 坐标，计算点对之间的距离并排序"""
        # 确定经纬度范围
        coords = np.array([[p['lat'], p['lon']] for p in points])
        min_lat, min_lon = np.min(coords, axis=0)
        max_lat, max_lon = np.max(coords, axis=0)
        self.logger.info(
            f"经纬度范围：纬度[{min_lat:.6f}, {max_lat:.6f}]，纬度范围[{max_lat-min_lat:.6f}]，"
            f"经度[{min_lon:.6f}, {max_lon:.6f}]，经度范围[{max_lon-min_lon:.6f}]")

        # 分配到网格
        grid_map = {}
        for img_info_dict in points:
            grid = self._assign_to_grid(
                img_info_dict['lat'], img_info_dict['lon'], grid_size, min_lat, min_lon)
            if grid not in grid_map:
                grid_map[grid] = []
            grid_map[grid].append(
                (img_info_dict['file'], img_info_dict['lat'], img_info_dict['lon']))
        
        self.logger.info(f"图像点已分配到 {len(grid_map)} 个网格中")

        # 在每个网格中计算两两距离并排序
        sorted_distances = {}
        for grid, images in grid_map.items():
            distances = []
            for (img1, lat1, lon1), (img2, lat2, lon2) in combinations(images, 2):
                dist = self._haversine(lat1, lon1, lat2, lon2)
                distances.append((img1, img2, dist))
            distances.sort(key=lambda x: x[2])  # 按距离升序排序
            sorted_distances[grid] = distances
            self.logger.debug(f"网格 {grid} 中计算了 {len(distances)} 个距离对")

        return sorted_distances

    def filter_dense_points(self, points, grid_size=0.001, distance_threshold=13):
        """过滤密集点，根据提供的距离阈值"""
        self.logger.info(f"开始过滤密集点 (网格大小: {grid_size}, 距离阈值: {distance_threshold}米)")
        
        # 获取每个网格中的图片的两两距离信息
        sorted_distances = self._get_distances(points, grid_size)

        to_del_imgs = []
        """遍历每个网格，删除网格中距离小于阈值的点"""
        for grid, distances in sorted_distances.items():
            grid_del_count = 0
            while distances:
                candidate_img1, candidate_img2, dist = distances[0]
                if dist < distance_threshold:
                    # 有小于阈值的距离，先删除最近的距离，这两个点作为候选点，要删掉一个
                    distances.pop(0)

                    # 获取候选图片1和图片2倒数第二短的距离
                    candidate_img1_dist = None
                    candidate_img2_dist = None
                    for distance in distances:
                        if candidate_img1 in distance:
                            candidate_img1_dist = distance[2]
                            break
                    for distance in distances:
                        if candidate_img2 in distance:
                            candidate_img2_dist = distance[2]
                            break

                    # 谁短删掉谁
                    if candidate_img1_dist and candidate_img2_dist:
                        if candidate_img1_dist < candidate_img2_dist:
                            to_del_img = candidate_img1
                        else:
                            to_del_img = candidate_img2
                        to_del_imgs.append(to_del_img)
                        grid_del_count += 1
                        self.logger.debug(f"在网格 {grid} 中删除密集点: {to_del_img} (距离: {dist:.2f}米)")
                        # 从距离列表中删除与被删除图片相关的记录
                        distances = [
                            distance for distance in distances if to_del_img not in distance]
                else:
                    break
            if grid_del_count > 0:
                self.logger.info(f"网格 {grid} 中删除了 {grid_del_count} 个密集点")

        # 过滤掉删除的图片，写入日志
        with open(self.log_file, 'a', encoding='utf-8') as f:
            for img in to_del_imgs:
                f.write(img+'\n')
        
        filtered_points = [point for point in points if point['file'] not in to_del_imgs]
        self.logger.info(f"密集点过滤完成，共删除 {len(to_del_imgs)} 个点，剩余 {len(filtered_points)} 个点")
        return filtered_points

    def filter_isolated_points(self, points, threshold_distance=0.001, min_neighbors=6):
        """过滤孤立点"""
        self.logger.info(f"开始过滤孤立点 (距离阈值: {threshold_distance}, 最小邻居数: {min_neighbors})")
        
        coords = np.array([[p['lat'], p['lon']] for p in points])
        kdtree = KDTree(coords)
        neighbors_count = [len(kdtree.query_ball_point(
            coord, threshold_distance)) for coord in coords]
        
        isolated_points = []
        with open(self.log_file, 'a', encoding='utf-8') as f:
            for i, p in enumerate(points):
                if neighbors_count[i] < min_neighbors:
                    isolated_points.append(p['file'])
                    f.write(p['file']+'\n')
                    self.logger.debug(f"删除孤立点: {p['file']} (邻居数: {neighbors_count[i]})")
            f.write('\n')
        
        filtered_points = [p for i, p in enumerate(points) if neighbors_count[i] >= min_neighbors]
        self.logger.info(f"孤立点过滤完成，共删除 {len(isolated_points)} 个点，剩余 {len(filtered_points)} 个点")
        return filtered_points
